-- keep wowace alive, donate!
local L = LibStub("AceLocale-3.0"):GetLocale("SpellReminder")

SpellReminder = LibStub('AceAddon-3.0'):NewAddon("SpellReminder", "AceConsole-3.0", "AceEvent-3.0", "AceBucket-3.0", "AceTimer-3.0")
SpellReminder.sm = LibStub("LibSharedMedia-3.0")

local tooltip = CreateFrame("GameTooltip","SpellReminderTooltipScanning")
tooltip:SetOwner( WorldFrame, "ANCHOR_NONE" );
tooltip.LL, tooltip.LR = {}, {}
do
	local i
	for i=1,30 do
		tooltip.LL[i],tooltip.LR[i] = tooltip:CreateFontString(),tooltip:CreateFontString()
		tooltip:AddFontStrings(tooltip.LL[i],tooltip.LR[i]);
	end
	function tooltip:FindTipText(match,noleft,noright)
		for i=1,self:NumLines() do
			local l,r = self.LL[i]:GetText() or "" , self.LR[i]:GetText() or ""
			start,fin,txt = string.find(l,match)
			if start then return start,fin,txt end
		end
	end
	function tooltip:GetFullTipText(noleft,noright)
		local fulltext = ""
		for i=1,self:NumLines() do
			local l,r = self.LL[i]:GetText() or "" , self.LR[i]:GetText() or ""
			if not noleft then fulltext = fulltext.." "..l end
			if not noright then fulltext = fulltext.." "..r end
		end
		return fulltext
	end
end

local function sortFunc(a, b)
	local toReturn
	
	if a.isTimer ~= b.isTimer then
		toReturn = a.isTimer
	elseif a.value == b.value then
		toReturn = a.name > b.name
	else
		toReturn = a.value > b.value
	end
	
	local eithergrp = a:GetGroup() or b:GetGroup() -- both bars are not always set.. for some reason.
	if eithergrp then
		for i,anchor in pairs(SpellReminder.anchors) do
			if anchor.LBframe == eithergrp and SpellReminder.db.profile.bargroups[i].reversesort then
				return not toReturn
			end
		end
	end
	return toReturn
end
local function ClipNumber(number, minimum, maximum, trueMax)
	return min(max(min(number,maximum),minimum),trueMax)
end
-- Thanks EBB
local roman_to_arabic = setmetatable({I = 1, V = 5, X = 10, L = 50, C = 100, D = 500, M = 1000}, {__index=function(self, roman)
	local arabic = 0
	local maxval = 0
	for i = roman:len(), 1, -1 do
		local digitval = self[roman:sub(i,i)]
		if digitval < maxval then
			arabic = arabic - digitval
		else
			arabic = arabic + digitval
			maxval = digitval
		end
	end
	self[roman] = arabic
	return arabic
end})

local _G = _G
local pairs = _G.pairs
local ipairs = _G.ipairs
local select = _G.select
local GetTime = _G.GetTime
local UnitName = _G.UnitName
local UnitGUID = _G.UnitGUID
local UnitBuff = _G.UnitBuff
local UnitDebuff = _G.UnitDebuff
local UnitIsDead = _G.UnitIsDead
local UnitExists = _G.UnitExists

local SPELLTYPE_NORMAL = ""
local SPELLTYPE_COOLDOWN = "CD"

local dbDefaultsProfile = {
	profile = {
		minimap = {
			hide = false,
		},
		ttg			= 0.3,
		disableEmphasis = false,
		lockbars	= false,
		debugmode	= false,
		docooldowns = true,
		-- output
		textWarnings	= L["Both"],
		audioWarnings	= L["None"],
		-- formatting
		textFormat	= "%t (%a) - %n",
		texture		= "Blizzard",
		baroverride	= false,
		barfont		= nil,
		barfontsize	= 11,
		barwidth	= 200,
		barheight	= 16,
		flashbars	= true,
		iconBorder	= true,
		-- spellsetup
		disableDetection	= false,
		defaultBarColor		= {["r"] = 1, ["g"] = 0, ["b"] = 0, ["a"] = 1},
		-- groups
		bargroups = {},
		--filters
		filterPrio = {'self','target','focus','buff','debuff','cooldown'},
		targetGroup_self	= 'normal',
		targetGroup_buff	= 'normal',
		targetGroup_debuff	= 'normal',
		targetGroup_cooldown= 'normal',
		targetGroup_focus	= 'normal',
		targetGroup_target	= 'normal',
		-- sink
		sinkOptions = {},
	},
}
local dbDefaultsChar = {
	char = {
		spellSetup	= {
			['*'] = {
				installed	= false,
				showSelf	= true,
				showOther	= true,
				singleTrack	= false,
				--fade		= 0,
			}
		},
	},
}

SpellReminder.hideWithoutStandby = true
SpellReminder.hasNoColor = true
SpellReminder.hasIcon = "Interface\\Icons\\Ability_Warrior_BattleShout"
SpellReminder.independentProfile = true

--SpellReminder.RegisteredBarGroups = {}

local spellSuffixes = {"",L[" Totem"],L[" Weapon"]}
--SpellReminder.barTypes = {"buff","debuff","cooldown"}

SpellReminder.anchors = {}

local defaultBarGroup = { name="New Bar", growup = false, emphasise=true, position = {}, scale = 1, }
--local castFailBuffer = 1
--local barRefreshBuffer = 0.5

-- stores information PER SPELL[Target]
SpellReminder.track = {
	casting = {},
	active = {},
}
local spellTracker = SpellReminder.track

function SpellReminder:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("SpellReminderDB",dbDefaultsProfile,true)
	self.chardb = LibStub("AceDB-3.0"):New("SpellReminderDBPerChar",dbDefaultsChar,true)
	if (self.chardb.global) then
		self:tmerge(self.chardb.global,self.chardb.char)
		self.chardb.global = nil
	end
	-- update bar idents
	for k,v in ipairs(self.db.profile.bargroups) do
		if v.ident ~= nil then
			self.db.profile.bargroups[v.ident] = v
			self.db.profile.bargroups[k] = nil
		end
	end

	LibStub("LibBars-1.0"):Embed(self)
	LibStub("LibSink-2.0"):Embed(self)
	self:SetSinkStorage(self.db.profile.sinkOptions)
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpellReminder", self.options, {'spellreminder','sr'})
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpellReminder", "SpellReminder")
	
	local ldb = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("SpellReminder", {
		type = "launcher",
		text = "SpellReminder",
		icon = "Interface\\Icons\\Ability_Warrior_BattleShout",
		OnClick = function(self, button)
			LibStub("AceConfigDialog-3.0"):Open("SpellReminder")
		end,
	})
	LibStub("LibDBIcon-1.0"):Register("SpellReminder", ldb, SpellReminder.db.ldb)
end

function SpellReminder:COMBAT_LOG_EVENT(...)
	local eventName = select(3,...)
	if self.combatlog_events[eventName] ~= nil then
		self.combatlog_events[eventName](self,self:ProcessCombatLogArgs(...))
	end
end

function SpellReminder:AutoTanks(post)
	local groupHP, maxHP, minHP, groupCount, inRaid = 0,0,0,0,UnitInRaid("player")
	if inRaid then
		for i = 1, 40, 1 do
			if UnitExists("raid"..i) then
				local unitHP = UnitHealthMax("raid"..i)
				groupHP = groupHP + unitHP
				maxHP = max(maxHP,unitHP)
				minHP = min(minHP,unitHP)
				groupCount = groupCount + 1
			end
		end
	else
		local unitHP = UnitHealthMax("player")
		groupHP = groupHP + unitHP
		maxHP = max(maxHP,unitHP)
		minHP = min(minHP,unitHP)
		groupCount = groupCount + 1
		for i = 1, 4, 1 do
			if UnitExists("party"..i) then
				local unitHP = UnitHealthMax("party"..i)
				groupHP = groupHP + unitHP
				maxHP = max(maxHP,unitHP)
				minHP = min(minHP,unitHP)
				groupCount = groupCount + 1
			end
		end
	end
	local averageHP = groupHP / groupCount
	local threshold = averageHP + ((maxHP-averageHP) / 3)
	
	if inRaid then
		for i = 1, 40, 1 do
			if UnitExists("raid"..i) then
				local unitHP = UnitHealthMax("raid"..i)
				if unitHP >= threshold then
					self:Print("AutoTank: "..UnitName("raid"..i))
				end
			end
		end
	else
		local unitHP = UnitHealthMax("player")
		if unitHP >= threshold then
			if post then
						
			else
				self:Print("AutoTank: "..UnitName("player").." "..unitHP.." >= "..threshold)
			end
		end	
		for i = 1, 4, 1 do
			if UnitExists("party"..i) then
				local unitHP = UnitHealthMax("party"..i)
				if unitHP >= threshold then
					self:Print("AutoTank: "..UnitName("party"..i).." "..unitHP.." >= "..threshold)
				end
			end
		end
	end
end

function SpellReminder:OnEnable()
	self:OnProfileEnable()
    -- Called when the addon is loaded
	self:RegisterEvent("PLAYER_LOGIN")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED","COMBAT_LOG_EVENT")
--	self:RegisterEvent("UNIT_SPELLCAST_SENT")
--	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
--	self:RegisterEvent("UNIT_SPELLCAST_STOP")
--	self:RegisterEvent("UNIT_SPELLMISS")
--	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
	self:RegisterEvent('UNIT_EXITED_VEHICLE',function()
		for e,castingInfo in pairs(spellTracker.active) do
			if castingInfo.tGUID == "VEHICLE" then
				self:CleanUp(eventName)--self:OnDurationEnded(eventName,true)
			end
		end
	end)
	self:RegisterEvent("PLAYER_DEAD")
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN","ProcessSpellCooldowns")
	self:RegisterEvent("PLAYER_ENTERING_WORLD",function() self:ProcessSpellCooldowns(true) end)
	
	self:RegisterEvent("UNIT_PET",function (unit)
		if unit ~= "player" then return end -- only process for players pet.
		for eventName,info in pairs(spellTracker.active) do -- first clear all CD timers, this will stop  multiple bars when zoning
			if info.tGUID == "PET" then
				self:CleanUp(eventName)--self:OnDurationEnded(eventName,true)
			end
		end
		self:ProcessAuras("pet")
	end)

	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP",function(unit,spell,rank)
		local info = self.chardb.char.spellSetup[spell]
		if unit == "player" and info.installed and info.persist ~= false then
			self:CleanUp(self:GetEventName(spell,UnitGUID("player")))
		end
	end)
	
	-- START
	self:RegisterCombatLogEvent('SPELL_CAST_SUCCESS')
	self:RegisterCombatLogEvent('UNIT_DIED')
	self:RegisterCombatLogEvent('PARTY_KILL','UNIT_DIED')
	self:RegisterCombatLogEvent('UNIT_DESTROYED','UNIT_DIED')
	self:RegisterEvent('UNIT_INVENTORY_CHANGED',function(unit) if unit == "player" then self:ScheduleTimer('ProcessAuras',0.5,"player")  end end)

	self:RegisterEvent("PLAYER_TARGET_CHANGED",function() self:ProcessAuras("target") end)
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT",function() self:ProcessAuras("mouseover") end)

	self:RegisterBucketEvent("UNIT_AURA",0.25,function(units)
		for unit in pairs(units) do
			self:ProcessAuras(unit,true)
		end
	end)

	self:RegisterEvent("PLAYER_TOTEM_UPDATE", function() self:ProcessAuras("player") end)
	self:RegisterEvent("PLAYER_FOCUS_CHANGED",function() self:ProcessAuras("focus")	end)
	
	self:ScheduleRepeatingTimer('RefreshAllBars',0.2)
	
	self:CreateAnchors()

	self:RefreshSpellOptions()
end

function SpellReminder:OnProfileEnable()
	-- convert old info
	self.db.profile.barposition = nil
	self.db.profile.splitSelf = nil
	
	if self.chardb.char.bargroups ~= nil then -- move bars from char to profile
		if self.db.profile.bargroups == nil then self.db.profile.bargroups = {} end
		self:tcopy(SpellReminder.chardb.char.bargroups,SpellReminder.db.profile.bargroups)
		self.chardb.char.bargroups = nil
	end
	-- end
	
	self:CreateAnchors()
	self:RefreshFilterPrioOptions()
end

function SpellReminder:GetEventName(spellName,targetName)
	if self.chardb.char.spellSetup[spellName] and self.chardb.char.spellSetup[spellName].installed and self.chardb.char.spellSetup[spellName].singleTrack == true then targetName = "" end
	return "SpellReminder_TRACKING_"..spellName..";"..targetName
end

function SpellReminder:GetCastingObject(spellName,displayName,spellRank,targetName,targetGUID,timeStarted,duration,remainder,eventName,colour,bartype,icon,targetID)
	targetID = self:GetTargetId(targetID)
--	local info = self.chardb.char.spellSetup[spellName]
	
	castingInfo = {}
	castingInfo.n 			= spellName
	castingInfo.displayName	= displayName
	castingInfo.rank		= spellRank
	castingInfo.t			= targetName
	castingInfo.tID			= targetID
	castingInfo.tGUID		= targetGUID
	castingInfo.s			= timeStarted
	castingInfo.d			= duration
	castingInfo.r			= remainder or ClipNumber(math.ceil(duration*0.125),5,15,duration)
	castingInfo.eventName	= eventName
	castingInfo.c			= colour
	castingInfo.barType		= bartype
	castingInfo.icon		= icon
--	castingInfo.persist		= info and info.persist ~= false
	return castingInfo
end
SpellReminder.idPriority = {"target","focus"}
function SpellReminder:GetTargetId(newID,oldID)
	-- prio :    focus -> target -> nil
	for i,v in ipairs(SpellReminder.idPriority) do
		if (oldID == nil and newID == v) or oldID == v or newID == v then return v end
	end
	return nil
end

function SpellReminder:GetSpellSetup(spellName,spellType)
	if spellType == SPELLTYPE_NORMAL then
		spellType = ''
	else
		spellType = spellType..';'
	end
	return self.chardb.char.spellSetup[spellType..spellName]
end

function SpellReminder:PLAYER_LOGIN()
	self:ProcessAuras("player")
	self:ProcessAuras("pet")
	self:ProcessSpellCooldowns()
end

function SpellReminder:OnDisable()

end

function SpellReminder:AddNewAnchor(overrideProperties)
	local newGroup = {}; self:tcopy(defaultBarGroup,newGroup)
	if overrideProperties == nil then return newGroup end
	-- need to override
	for k,v in pairs(overrideProperties) do newGroup[k] = v end
	
	if newGroup.ident == nil then newGroup.ident = newGroup.name end
	
	--table.insert(SpellReminder.db.profile.bargroups,newGroup)
	self.db.profile.bargroups[newGroup.ident] = newGroup

	self:CreateAnchors()
	
	return newGroup
end

--[[function SpellReminder:GetAnchorByIdent(val)
	for k,anchor in pairs(SpellReminder.db.profile.bargroups) do
		if anchor.ident == val then return anchor end
	end
	return nil
end]]

function SpellReminder:AnchorExists(ident)
	if SpellReminder.db.profile.bargroups[ident] == nil then return false end
	return true
end

function SpellReminder:CreateAnchors()
	if not self:AnchorExists("normal") then
		self:AddNewAnchor({ ident="normal",		name=L["Normal"],	position=(self.db.profile.barposition and self.db.profile.barposition["normal"]) or {}, })
		return
	end
	if not self:AnchorExists("emphasis") then
		self:AddNewAnchor({ ident="emphasis",	name=L["Emphasis"],	position=(self.db.profile.barposition and self.db.profile.barposition["emphasis"]) or {}, })
		return
	end

	for i,groupinfo in pairs(SpellReminder.db.profile.bargroups) do
		if groupinfo == nil then
			self.db.profile.bargroups[i] = nil
		else
			if self.anchors[groupinfo.ident] == nil then self.anchors[groupinfo.ident] = {} end
			self:tcopy(groupinfo,self.anchors[groupinfo.ident])
			-- new LibBars
			if self.anchors[groupinfo.ident].LBframe == nil then
				self.anchors[groupinfo.ident].LBframe = self:NewBarGroup(groupinfo.name, 1, self.db.profile.barwidth, self.db.profile.barheight, "ReminderFrame_"..i)
				self.anchors[groupinfo.ident].LBframe.RegisterCallback(self,'AnchorMoved')
			end
			local LBframe = self.anchors[groupinfo.ident].LBframe
			LBframe:SetWidth(self.db.profile.barwidth)
			LBframe.width = self.db.profile.barwidth
			LBframe:SetHeight(self.db.profile.barheight)
			LBframe.height = self.db.profile.barheight
			LBframe:SetSortFunction(sortFunc)
			LBframe:ClearAllPoints()
			if self.db.profile.bargroups[i] and self.db.profile.bargroups[i].position and self.db.profile.bargroups[i].position.x and self.db.profile.bargroups[i].position.y and self.db.profile.bargroups[i].position.point and self.db.profile.bargroups[i].position.anchor then
				LBframe:SetPoint(self.db.profile.bargroups[i].position.point, UIParent, self.db.profile.bargroups[i].position.anchor, self.db.profile.bargroups[i].position.x, self.db.profile.bargroups[i].position.y)
			else
				LBframe:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
			end
			LBframe:SetScale(groupinfo.scale or 1)
			LBframe:ReverseGrowth((self.db.profile.bargroups[i] and self.db.profile.bargroups[i].growup) or false)
		end
	end
	self:RefreshBarGroupList()
	self:RefreshGroupOptions()
	self:RefreshSpellOptions()
	
	self:ToggleAnchors()
end

function SpellReminder:ResetAnchors()
	if not self.db.profile.lockbars then self:ToggleAnchors() end
	for idx,anchor in pairs(self.anchors) do
		self:ResetAnchor(idx)
	end
	
	self:SaveAnchorPositions()
end

function SpellReminder:ResetAnchor(id)
	local anchor = self.anchors[id]
	anchor.LBframe:ClearAllPoints()
	anchor.LBframe:SetPoint("CENTER", 0, 50 / anchor.LBframe:GetScale())
end

function SpellReminder:ToggleAnchors()
	if self.db.profile.lockbars then
		for _,anchor in pairs(self.anchors) do
			anchor.LBframe:HideAnchor()
		end
	else
		for _,anchor in pairs(self.anchors) do
			anchor.LBframe:ShowAnchor()
		end
	end
end

function SpellReminder:AnchorMoved(cbk, group, x, y)
	self:SaveAnchorPositions()
end

function SpellReminder:DeleteBarGroup(id)
	self.anchors[id].LBframe:Hide()
	self.anchors[id] = nil
	self.db.profile.bargroups[id] = nil
	
	local normalId = 'normal'
	
	for spellName,info in pairs(self.chardb.char.spellSetup) do
		if info.allowCustomGroup == true and info.customGroup == id then
			info.allowCustomGroup = false
			info.customGroup = normalId
			self:Print(L["%s has had its custom group reset."]:format(spellName))
		end
	end
	
	if self.db.profile.targetGroup_self		== id then self.db.profile.targetGroup_self = normalId; 	self:Print(L["Filter for '%s' has been reset."]:format("Self")) end
	if self.db.profile.targetGroup_buff		== id then self.db.profile.targetGroup_buff = normalId; 	self:Print(L["Filter for '%s' has been reset."]:format("Buffs")) end
	if self.db.profile.targetGroup_debuff	== id then self.db.profile.targetGroup_debuff = normalId; 	self:Print(L["Filter for '%s' has been reset."]:format("Debuffs")) end
	if self.db.profile.targetGroup_cooldown	== id then self.db.profile.targetGroup_cooldown = normalId; self:Print(L["Filter for '%s' has been reset."]:format("Cooldowns")) end
	if self.db.profile.targetGroup_focus	== id then self.db.profile.targetGroup_focus = normalId; 	self:Print(L["Filter for '%s' has been reset."]:format("Focus")) end
	if self.db.profile.targetGroup_target	== id then self.db.profile.targetGroup_target = normalId;	self:Print(L["Filter for '%s' has been reset."]:format("Target")) end
	
	self:Print(L["You should reload your UI after deleting a bar group."])
	
	self:RefreshSpellOptions()
	self:CreateAnchors()
end

function SpellReminder:SaveAnchorPositions()
	for i, anchorFrame in pairs(self.anchors) do
		local point, _, anchor, x, y = anchorFrame.LBframe:GetPoint()
		self.db.profile.bargroups[i].position.x = x --* s
		self.db.profile.bargroups[i].position.y = y --* s
		self.db.profile.bargroups[i].position.anchor = anchor
		self.db.profile.bargroups[i].position.point = point
	end
end

function SpellReminder:CreateAnchor(name, text, cRed, cGreen, cBlue)
	local f = CreateFrame("Button", "SpellReminder_"..name.."Anchor", UIParent)
	f:SetWidth(200)
	f:SetHeight(25)

	f.owner = self

	f:ClearAllPoints()
	
	if self.db.profile.bargroups[name] and self.db.profile.bargroups[name].position and self.db.profile.bargroups[name].position.x and self.db.profile.bargroups[name].position.y and self.db.profile.bargroups[name].position.point and self.db.profile.bargroups[name].position.anchor then
		f:SetPoint(self.db.profile.bargroups[name].position.point, UIParent, self.db.profile.bargroups[name].position.anchor, self.db.profile.bargroups[name].position.x, self.db.profile.bargroups[name].position.y)
	else
		f:SetPoint("CENTER", UIParent, "CENTER", 0, 50)--/s)
	end

	f:SetScript("OnDragStart", function() this:StartMoving() end )
	f:SetScript("OnDragStop",
		function()
			this:StopMovingOrSizing()
			self:SaveAnchorPositions()
        end)

	f:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background",
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
                                            tile = false, tileSize = 16, edgeSize = 16,
                                            insets = { left = 5, right =5, top = 5, bottom = 5 }})
	f:SetBackdropColor(cRed,cGreen,cBlue,.6)
	f:SetMovable(true)
	f:RegisterForDrag("LeftButton")

	f.Text = f:CreateFontString(nil, "OVERLAY")
	f.Text:SetFontObject(GameFontNormalSmall)
	f.Text:ClearAllPoints()
	f.Text:SetTextColor(1, 1, 1, 1)
	f.Text:SetWidth(200)
	f.Text:SetHeight(25)
	f.Text:SetPoint("TOPLEFT", f, "TOPLEFT")
	f.Text:SetJustifyH("CENTER")
	f.Text:SetJustifyV("MIDDLE")
	f.Text:SetText(text)

	f:Hide()

	return f
end

function SpellReminder:CreateTestBars()
	local i
	for i=1, 5 do
		local castingInfo
		if math.random()>0.5 then -- debuff
			castingInfo = self:GetCastingObject("Test "..i,nil,nil,"Enemy "..i,'NaN',GetTime(),i*5,nil,"TEST_MANUAL_"..i,{r=math.random(),g=math.random(),b=math.random()},"debuff",'Interface\\Icons\\INV_Misc_QuestionMark')
		else
			local pseudoTarget = (math.random() > 0.5 and "Player "..i) or "Friendly "..i
			castingInfo = self:GetCastingObject("Test "..i,nil,nil,pseudoTarget,'NaN',GetTime(),i*5,nil,"TEST_MANUAL_"..i,{r=math.random(),g=math.random(),b=math.random()},"buff",'Interface\\Icons\\INV_Misc_QuestionMark')
		end
		castingInfo.test = true
		self:TrackSpell(castingInfo)
	end

	local castingInfo = self:GetCastingObject("Focus Test",nil,nil,"Focus",'NaN',GetTime(),30,nil,"TEST_FOCUS",{r=math.random(),g=math.random(),b=math.random()},"buff",'Interface\\Icons\\INV_Misc_QuestionMark',"focus")
	castingInfo.test = true
	self:TrackSpell(castingInfo)

	local castingInfo = self:GetCastingObject("Target Test",nil,nil,"Target",'NaN',GetTime(),30,nil,"TEST_TARGET",{r=math.random(),g=math.random(),b=math.random()},"buff",'Interface\\Icons\\INV_Misc_QuestionMark',"target")
	castingInfo.test = true
	self:TrackSpell(castingInfo)
	
	-- must reset options or AceConfig crashes on second execute
	self:RefreshBarGroupList()
	self:RefreshGroupOptions()
	self:RefreshSpellOptions()
end

function SpellReminder:ClearSpells(clearAll)
	for spellName,info in pairs(self.chardb.char.spellSetup) do
		if (clearAll == true) or (clearAll == "inactive" and info.active == false) then
			self:ClearSpell(spellName,true)
		elseif (info.detected ~= true) then
			local found = false
			local i = 1
			while true do
				local spell, rank = GetSpellBookItemName(i, BOOKTYPE_SPELL)
				if (not spell) then	break end
				if spell == spellName then
					found = true
				end
				i = i + 1
			end
			
			if found == false then self:ClearSpell(spellName,true) end
		end
	end
	
	self:RefreshSpellOptions()
end

function SpellReminder:ClearSpell(spellName,dontRefresh)
	self.chardb.char.spellSetup[spellName] = nil
	self.spellOptions[spellName] = nil
	if not dontRefresh then self:RefreshSpellOptions() end
end

function SpellReminder:InstallWatch(spellName,rank,reminder,duration,icon)--,sharedName,reminder)
	if spellName == nil then return end
	if duration == nil then return end -- this can happen if the spell isnt available on this character

	local realSpellName = spellName
	if string.find(realSpellName,":") then
		_,_,_,realSpellName = string.find(realSpellName,"(.*):(.*)")
	end
		
	self:DebugPrint("Adding spell: "..realSpellName)
	self.chardb.char.spellSetup[spellName].installed = true
	if (self.chardb.char.spellSetup[spellName].active == nil) then
		self.chardb.char.spellSetup[spellName].active = true
	end
	self.chardb.char.spellSetup[spellName].displayName = self.chardb.char.spellSetup[spellName].displayName or realSpellName
	self.chardb.char.spellSetup[spellName].rank		= rank
	
	self.chardb.char.spellSetup[spellName].icon = icon or self:GetTexture(spellName)
	
	if self.chardb.char.spellSetup[spellName].d == nil or duration > self.chardb.char.spellSetup[spellName].d then
		self.chardb.char.spellSetup[spellName].d	= duration
		local defaultReminder = ClipNumber(math.ceil(duration*0.125),5,15,duration)--max(min(math.ceil(duration*0.125),15),5)
		self.chardb.char.spellSetup[spellName].r	= self.chardb.char.spellSetup[spellName].r or reminder or defaultReminder

	end
	
	self:RefreshSpellOptions()
	return self.chardb.char.spellSetup[spellName]
end

function SpellReminder:FlagIsSet(flags,flag)
	return bit.band(flags,flag) == flag
end

function SpellReminder:SPELL_CAST_SUCCESS(t)
	if not (self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_TYPE_PLAYER) and self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_AFFILIATION_MINE)) then return end
	if self:FlagIsSet(t.destFlags,COMBATLOG_OBJECT_NONE) then
		t.destGUID = t.sourceGUID
		t.destName = t.sourceName
		t.destFlags = t.sourceFlags
	end
	
	self:DebugPrint(string.format("SPELL_CAST_SUCCESS: %s -> %s : %s (%s~%s)",t.sourceName or '',t.destName or '',t.spellName or '',t.spellId or '',t.spellSchool or 0))
	local _,spellRank, icon = GetSpellInfo(t.spellId)
	
	if (not self.chardb.char.spellSetup[t.spellName] or not self.chardb.char.spellSetup[t.spellName].installed) then
		if self.db.profile.disableDetection == false and (IsControlKeyDown() and IsShiftKeyDown()) then
			_,_,texture = GetSpellInfo(t.spellId)
			self:InstallWatch(t.spellName,"",nil,self:GetDuration(t.spellName),texture)
		else
			return
		end
	end
	local info = self.chardb.char.spellSetup[t.spellName]
	info.spellId = t.spellId
	if not info.installed or not info.active or not info.d or info.persist == false then return end

	local duration = info.d
	
	if spellTracker.casting == nil then spellTracker.casting = {} end
	local bt = 'buff'
	if self:FlagIsSet(t.destFlags,COMBATLOG_OBJECT_REACTION_HOSTILE) then
		bt = 'debuff'
		self:SetSpellCategory(t.spellName,'Debuffs')
	else
		self:SetSpellCategory(t.spellName,'Buffs')
	end
	
	--local bt = ((destName == nil or self:FlagIsSet(destFlags,COMBATLOG_OBJECT_REACTION_HOSTILE)) and "buff") or "debuff" --bartype
	local eventName, castingEvent = self:GetEventName(t.spellName,t.destGUID)
	self:TrackSpell(self:GetCastingObject(t.spellName,info.displayName,spellRank,t.destName,t.destGUID,GetTime(),duration,info.r,eventName,info.c,bt,info.icon))
end
function SpellReminder:SPELL_MISSED(t)
	if not (self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_TYPE_PLAYER) and self:FlagIsSet(t.sourceFlags,COMBATLOG_OBJECT_AFFILIATION_MINE)) then return end
	local _,spellRank = GetSpellInfo(t.spellId)
	local eventName = self:GetEventName(t.spellName,t.destGUID)
--	for eventName,spell in pairs(spellTracker.casting) do
	--	if spell.n == t.spellName and spell.rank == spellRank then
			spellTracker.casting[eventName] = nil
--		end
--	end
--	for eventName,spell in pairs(spellTracker.active) do
--		if spell.n == t.spellName and spell.rank == spellRank then
--			self:OnDurationEnded(eventName,true)
--		end
--	end
end
function SpellReminder:UNIT_DIED(t)
	self:DebugPrint(t.event..": "..t.destName)
	for eventName,info in pairs(spellTracker.casting) do
		if info.tGUID == t.destGUID then
			self:CleanUp(eventName)
		end
	end
	for eventName,info in pairs(spellTracker.active) do
		if info.tGUID == t.destGUID then
			self:CleanUp(eventName)
		end
	end
end

function SpellReminder:FinalCastSuccess(eventName)
	if spellTracker.casting[eventName] == nil then return end
	
	for e,info in pairs(spellTracker.active) do -- check if pet also has this buff, if it does, cancel the spellcast
		if info.tGUID == "PET" then -- spells on pet are always given the guid PET, snce dismissing the pet always issues a new GUID
			self:CleanUp(eventName)
			return
		end
	end

	self:DebugPrint("finalcast: "..spellTracker.casting[eventName].n)

	-- will be picked up by the aura processor
	local playerGUID = UnitGUID("player")
	local buffDuration = spellTracker.casting[eventName].d
	if buffDuration == nil then return end

	self:TrackSpell(spellTracker.casting[eventName])
	spellTracker.casting[eventName] = nil
end

function SpellReminder:IsTracking(eventName)
	return spellTracker.active[eventName] ~= nil
end

function SpellReminder:TrackSpell(castingInfo,dontRefresh)
	if type(castingInfo) ~= "table" then error("TrackSpell failed.  Expected table, got nil") return end
	local eventName = castingInfo.eventName
	
	local setupInfo = self.chardb.char.spellSetup[castingInfo.n]
	if castingInfo.tGUID ~= "COOLDOWN" and setupInfo.installed then
		if (not setupInfo.showSelf and castingInfo.tGUID == UnitGUID("player")) or
		   (not setupInfo.showOther and castingInfo.tGUID ~= UnitGUID("player")) then
			spellTracker.active[eventName] = nil
			spellTracker.casting[eventName] = nil
			return
		end
	end -- this spell is flagged as Self Only, but its being cast on another target, ignore bar creation and free active and casting tables for this event.

	if (setupInfo.last_seen == nil or setupInfo.last_seen < (time() - (86400*7))) then
		SpellReminder:RefreshSpellOptions()
	end
	setupInfo.last_seen = time()
	if (spellTracker.active[eventName] == nil) then
		spellTracker.active[eventName] = {}
		self:DebugPrint("Tracking: "..castingInfo.n)
		-- Announce
		if setupInfo.announceCast then
		local spellName, spellTarget = castingInfo.displayName, (castingInfo.t and castingInfo.t ~= UnitName("player") and " ("..castingInfo.t..")") or ""
		local message = L["%s Has been cast.  Will expire in %d seconds!"]:format(spellName..spellTarget,math.floor(castingInfo.d))
			if setupInfo.announceSay then SendChatMessage(message, "SAY") end
			if setupInfo.announceYell then SendChatMessage(message, "YELL") end
			if setupInfo.announceParty then SendChatMessage(message, "PARTY") end
			if setupInfo.announceRaid then SendChatMessage(message, "RAID") end
		end
	end
		
	self:tcopy(castingInfo,spellTracker.active[eventName])
	castingInfo = nil

	return spellTracker.active[eventName]
end

function SpellReminder:RefreshAllBars()
	self:DebugPrint('RefreshAllBars')
	-- add new bars
	for e,castingInfo in pairs(spellTracker.active) do
		local setupInfo = self.chardb.char.spellSetup[castingInfo.n]
		local timeLeft = castingInfo.d - (GetTime() - castingInfo.s)
	--	local alreadyExists, needsRefresh = self.candybar:IsCandyBarRegistered(e), false
--		if alreadyExists then
--			local _,totTime,elapsed = self.candybar:CandyBarStatus(e)
--			local barTimeLeft = totTime - (elapsed or 9999)
--			needsRefresh = (barTimeLeft <= timeLeft-barRefreshBuffer) or (barTimeLeft >= timeLeft+barRefreshBuffer) or (castingInfo.applications ~= nil) -- time left is > 1sec out of sync (must be 1second, since time() does not return ms precision)
--		end
		-- persistant ??????
		if (timeLeft <= 0) or (castingInfo.test == false and not setupInfo.active) then
			self:OnDurationEnded(e)
		else--if castingInfo.bar == nil or castingInfo.bar.SRstartTime ~= castingInfo.s then-- (castingInfo.bar and castingInfo.bar.startTime ~= castingInfo.s) or forceRefresh then--if --[[(not alreadyExists) or ]]needsRefresh or (castingInfo.targetGroupId ~= targetGroup) or forceRefresh then
--			local resort = false
	--		if castingInfo.bar == nil or castingInfo.bar.SRstartTime ~= castingInfo.s then
	--			resort = true
	--		end
			if setupInfo.installed and setupInfo.c and setupInfo.c ~= castingInfo.c then castingInfo.c = {}; self:tcopy(setupInfo.c,castingInfo.c) end-- copy to prevent setup color from being recycled by reference
			self:CreateReminderBar(e,castingInfo,timeLeft)
			local reminder = castingInfo.r or SpellReminder.chardb.char.spellSetup[castingInfo.n]["r"] -- castingInfo.r
			if timeLeft - reminder <= 0 then
				self:OnDurationEnding(e)
			end
		end
	end
	
	-- any which are still existing as bars, but are no longer active?
	for i, anchor in pairs(self.anchors) do
		local bars = anchor.LBframe:GetBars()
		if bars then for e,bar in pairs(bars) do
			if spellTracker.active[e] == nil then --and not bar.fading then
				bar:Fade()
			end
		end end
	end
end

function SpellReminder:FindTargetGroup(castingInfo)
	local targetGroupId = nil
	local setupInfo = self.chardb.char.spellSetup[castingInfo.n]
	
	if setupInfo.installed and setupInfo.allowCustomGroup and setupInfo.customGroup then -- and castingInfo.barType ~= "cooldown" then
		return setupInfo.customGroup
	end
	
	for _,fltrName in ipairs(self.db.profile.filterPrio) do
		if (castingInfo.barType == fltrName) or castingInfo.tGUID == UnitGUID((fltrName == "self" and "player") or fltrName) then
			return self.db.profile['targetGroup_'..fltrName]
		end
	end
	return 'normal'
end

function SpellReminder:CreateReminderBar(e, castingInfo, timeLeft, targetGroupId)
	local color = castingInfo.c or self.db.profile.defaultBarColor
	local duration = castingInfo.d
	local reminder = castingInfo.r or SpellReminder.chardb.char.spellSetup[castingInfo.n]["r"]
	local target = castingInfo.t or UnitName("player")

	if targetGroupId == nil then
		targetGroupId = self:FindTargetGroup(castingInfo)
		if timeLeft <= reminder and (self.db.profile.bargroups[targetGroupId] and self.db.profile.bargroups[targetGroupId].emphasise) then
            targetGroupId = 'emphasis'
		end
	end
	if not targetGroupId then
		if castingInfo.bar then
			castingInfo.bar:Fade()
			castingInfo.bar = nil
		end
		return
	end
	
	castingInfo.targetGroupId = targetGroupId
	local targetLibBarGroup = self.anchors[targetGroupId].LBframe
	
	-- build name
	local text = castingInfo.customFormat or self.db.profile.textFormat
	text = string.gsub(text,"%%n",castingInfo.displayName or castingInfo.n)
	text = string.gsub(text,"%%r",castingInfo.rank or "")
	text = string.gsub(text,"%%t",target)
	if castingInfo.applications == 0 then castingInfo.applications = nil end
	text = string.gsub(text,"%%a",castingInfo.applications or "")
	text = string.gsub(text, "[(][)]", "") -- trim empty parenthesis
	text = string.gsub(text, "^%s*(.-)%s*$", "%1") -- trim whitespaces
	
	if castingInfo.bar and castingInfo.bar:GetGroup() and not targetLibBarGroup:GetBar(e) then
		local oldGroup = castingInfo.bar:GetGroup()
		oldGroup:MoveBarToGroup(castingInfo.bar,targetLibBarGroup)--RemoveBar(castingInfo.bar)
		oldGroup:SortBars()
		targetLibBarGroup:SortBars()
	end
	if (not castingInfo.bar) or (castingInfo.s ~= castingInfo.bar.SRstartTime) or (castingInfo.bar:GetLength() ~= self.db.profile.barwidth) or (castingInfo.bar:GetThickness() ~= self.db.profile.barheight) then
		targetLibBarGroup:SetTexture(self.sm:Fetch("statusbar", self.db.profile.texture))
		local bar, isNew = targetLibBarGroup:NewTimerBar(e, text, timeLeft, duration, castingInfo.icon, (self.db.profile.flashbars and reminder) or 0)
		if isNew then
			self:DebugPrint("Created Bar: "..castingInfo.n..": duration="..(duration or "no duration")..": reminder="..(reminder or "no reminder")..": timeleft="..(timeLeft or "none"))
		end
		if self.db.profile.barfont ~= nil then
			local font = SpellReminder.sm:Fetch('font', self.db.profile.barfont)
			bar:SetFont(font,self.db.profile.barfontsize)
		end
		bar.SRstartTime = castingInfo.s
		bar.updateDelay = 0 -- update on every frame .  Jerky bars = bad
		castingInfo.bar = bar
		targetLibBarGroup:SortBars()
	end
	castingInfo.bar:SetText(text)
	castingInfo.bar:SetIcon(castingInfo.icon)
	castingInfo.bar:UnsetAllColors()
	castingInfo.bar:SetColorAt(0,color.r,color.g,color.b,1)
	if self.db.profile.iconBorder then
		castingInfo.bar.icon:SetTexCoord(0, 1, 0, 1)
	else -- crop off the border
		castingInfo.bar.icon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
	end
end

function SpellReminder:PLAYER_DEAD()
	for eventName,info in pairs(spellTracker.casting) do
		if info ~= nil then
			self:CleanUp(eventName)
			--self:OnDurationEnded(eventName,true)
		end
	end
	for eventName,info in pairs(spellTracker.active) do
		if info ~= nil then
			self:CleanUp(eventName)--self:OnDurationEnded(eventName,true)
		end
	end
end

function SpellReminder:OnDurationEnding(eventName,quiet)
	local castingInfo = spellTracker.active[eventName]
	if not castingInfo then self:CleanUp(eventName) return end
	if castingInfo.endingProcessed == true then return end
	castingInfo.endingProcessed = true
	self:DebugPrint("OnDurationEnding: "..eventName)
	--if castingInfo.barType == "cooldown" then return end
	local spellID, spellName, spellTarget = castingInfo.n, castingInfo.displayName, (castingInfo.t and castingInfo.t ~= UnitName("player") and " ("..castingInfo.t..")") or ""
	local setupInfo = self.chardb.char.spellSetup[spellID]
		
	if not quiet and setupInfo and setupInfo.installed then
		local soundfile = self.db.profile.audioWarnings
		if setupInfo.installed and setupInfo.allowCustomSound and setupInfo.customSound then soundfile = setupInfo.customSound end
		PlaySoundFile(SpellReminder.sm:Fetch('sound', soundfile))
		
		local timeLeft = castingInfo.d - (GetTime() - castingInfo.s)
		local message = L["%s over in %d seconds!"]:format(spellName..spellTarget,math.floor(timeLeft))
	
		-- UI Text Warning
		if (self.db.profile.textWarnings == "Both" or self.db.profile.textWarnings == "Expiring Only") then
			self:Pour(message,1,1,1,nil,nil,nil,true)
		end

		-- Announce
		if setupInfo.announceEnding then
			if setupInfo.announceSay then SendChatMessage(message, "SAY") end
			if setupInfo.announceYell then SendChatMessage(message, "YELL") end
			if setupInfo.announceParty then SendChatMessage(message, "PARTY") end
			if setupInfo.announceRaid then SendChatMessage(message, "RAID") end
		end
	end
end

function SpellReminder:OnDurationEnded(eventName,quiet)
	self:DebugPrint("OnDurationEnded: "..eventName)
	
	local castingInfo = spellTracker.active[eventName]
--	if not castingInfo or castingInfo.barType == "cooldown" then self:CleanUp(eventName) return end
	local spellID, spellName, spellTarget = castingInfo.n, castingInfo.displayName, (castingInfo.t and castingInfo.t ~= UnitName("player") and " ("..castingInfo.t..")") or ""
	local setupInfo = self.chardb.char.spellSetup[spellID]

	if not quiet and setupInfo and setupInfo.installed then
		local soundfile = self.db.profile.audioWarnings
		if setupInfo.installed and setupInfo.allowCustomSound and setupInfo.customSound then soundfile = setupInfo.customSound end
		PlaySoundFile(SpellReminder.sm:Fetch('sound', soundfile))
		
		local message = L["%s has EXPIRED!"]:format(spellName..spellTarget)
		
		-- UI Text Warning
		if (self.db.profile.textWarnings == "Both" or self.db.profile.textWarnings == "Expired Only") then
			self:Pour(message,1,1,1,nil,nil,nil,true)
		end
		
		-- Announce
		if setupInfo.announceEnded then
			if setupInfo.announceSay then SendChatMessage(message, "SAY") end
			if setupInfo.announceYell then SendChatMessage(message, "YELL") end
			if setupInfo.announceParty then SendChatMessage(message, "PARTY") end
			if setupInfo.announceRaid then SendChatMessage(message, "RAID") end
		end
	end
	
	self:CleanUp(eventName)
end

function SpellReminder:CleanUp(eventName)
	if eventName == nil then return end

	spellTracker.casting[eventName] = nil
	spellTracker.active[eventName] = nil
end

--[[function SpellReminder:GetRegisteredGroup(eventName)
	local regdGroup = nil
	for k,v in pairs(self.anchors) do
		if self.candybar:IsCandyBarRegisteredWithGroup(eventName,"ReminderCandy_"..k) then
			regdGroup = k
			break
		end
	end
	return regdGroup
end]]

function SpellReminder:GetTexture(name)
	local i = 1
	while true do
		local nm = GetSpellBookItemName(i, BOOKTYPE_SPELL)
		if not nm then break
		elseif nm == name then
			return GetSpellTexture(i, BOOKTYPE_SPELL)
		end
		i = i + 1
	end
	-- still not found? try pets spellbook
	local i = 1
	while true do
		local nm = GetSpellBookItemName(i, BOOKTYPE_PET)
		if not nm then break
		elseif nm == name then
			return GetSpellTexture(i, BOOKTYPE_PET)
		end
		i = i + 1
	end
	
	if true then return "Interface\\Icons\\INV_Misc_QuestionMark" end -- we dont know the texture, use questionmark
	
	-- finally, do the talents
	local numTabs = GetNumTalentTabs();
	local t
	for t=1, numTabs do
	    local numTalents = GetNumTalents(t);
	    for i=1, numTalents do
	        local nameTalent, icon, tier, column, currRank, maxRank= GetTalentInfo(t,i);
			if nameTalent == spellName and currRank > 0 then
				local buff, debuff = self:FindDurationInText(tiptext)
				return buff, debuff, currRank
			end
	    end
	end
end

function SpellReminder:BarExists(spellName)
	for e, info in pairs(spellTracker.active) do
		if info.n == spellName then
			return true
		end
	end
	return false
end

function SpellReminder:GetPlayerBuffDuration(spellName)
	local i = 1
	while (UnitBuff("player",i) ~= nil) do
		local name, rank, iconTexture, count, duration, timeLeft =  UnitBuff("player", i)
		if name == spellName and duration ~= nil then return duration end
		i = i + 1
	end
	local i = 1
	while (UnitBuff("pet",i) ~= nil) do
		local name, rank, iconTexture, count, duration, timeLeft =  UnitBuff("pet", i)
		if name == spellName and duration ~= nil then return duration end
		i = i + 1
	end
end

function SpellReminder:IsInSpellbook(spellName)
	for _,suffix in ipairs(spellSuffixes) do
		local i = 1
		while true do
			local spell, rank = GetSpellBookItemName(i, BOOKTYPE_SPELL)
			if (not spell) then	break end
			if spell == spellName..suffix then
				return true
			end
			i = i + 1
		end
	end
	
	return false
end

function SpellReminder:GetTempBuffName(id)
	local rank
	tooltip:SetInventoryItem("player", id)
	local _, _, buffname = tooltip:FindTipText("^(.+) %(%d+ [^%)]+%)$")
	if buffname then
		buffname = string.gsub(buffname, " %(%d+ [^%)]+%)", "") -- remove 2nd brackets for buffs with charges
		local tname, trank = string.match(buffname, "(.*) (%d*)$")
		if tname then
			buffname = tname
			rank = trank
		else
			local tname, trank = string.match(buffname, "(.*) ([CDILMVX]*)$")
			if tname then
				buffname = tname
				rank = roman_to_arabic[trank]
			end
		end
		return buffname, rank
	end
	local itemlink = GetInventoryItemLink("player", id)
	if itemlink then
		local name = GetItemInfo(itemlink)
		return name or "Weapon "..id
	end
	return "Weapon "..id
end

function SpellReminder:GetDuration(spellName, chkRank)
	for _,suffix in pairs(spellSuffixes) do
		local toreturn = nil
		local lastRank = nil
		local i = 1
		while true do
			local spell, rank = GetSpellBookItemName(i, BOOKTYPE_SPELL)
			if (not spell) then	break end
			if spell == spellName..suffix then
				if (chkRank == nil and (lastRank == nil or rank > lastRank)) or (rank == chkRank) then
					lastRank = rank
					tooltip:SetSpellBookItem(i,BOOKTYPE_SPELL)
					local tiptext = tooltip:GetFullTipText()
					local buff = self:FindDurationInText(tiptext)
					toreturn = { buff, rank }
				end
			end
			i = i + 1
		end
		if toreturn ~= nil then return toreturn[1], toreturn[2] end
	end

	return nil
end

function SpellReminder:FindDurationInText(tiptext)
	local pre = {"for","over","lasts","lasting","up to"}
	local post = {"sec","min","hour"}
	local durationBuff = nil
	local durationDebuff = nil
	for _,preVal in ipairs(pre) do -- loop through prefixes
		preVal = string.lower(preVal)
		for _,postVal in ipairs(post) do  -- loop through suffixes
			postVal = string.lower(postVal)
			for newDur in string.gmatch(string.lower(tiptext),preVal.." (%d+) "..postVal) do
				newDur = tonumber(newDur)
				if newDur ~= nil then
					if postVal == "min" then
						newDur = newDur * 60
					elseif postVal == "hour" then
						newDur = newDur * 60 * 60
					end
					if (durationBuff == nil) then
						durationBuff = newDur
					elseif (newDur > durationBuff) then
						durationDebuff = durationBuff
						durationBuff = newDur
					end
				end
			end
		end
	end
	
	return tonumber(durationBuff), tonumber(durationDebuff)
end

function SpellReminder:DebugPrint(text)
	if not self.db.profile.debugmode then return end
	self:Print(text)
end

function SpellReminder:tcopy(from,to)   -- "to" must be a table (possibly empty)
	assert(type(from)=="table","Error copying table contents, source is not a table.")
	assert(type(to)  =="table","Error copying table contents, destination is not a table.")
	for k,v in pairs(from) do
		if(type(v)=="table") then
			if (type(to[k]~='table')) then to[k] = {} end
			self:tcopy(v,to[k])
		else
			to[k] = v
		end
	end
end

function SpellReminder:tmerge(from,to)   -- "to" must be a table (possibly empty)
	assert(type(from)=="table","Error copying table contents, source is not a table.")
	assert(type(to)  =="table","Error copying table contents, destination is not a table.")
	for k,v in pairs(from) do
		if(type(v)=="table") then
			if (type(to[k])~='table') then to[k] = {} end
			self:tmerge(v,to[k])
		else
			to[k] = v
		end
	end
end